# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import abc
import os
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import torch
import torch.nn as nn
import torchmetrics
from pytorch_lightning import LightningModule
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.data.datasets import _TransformedDataset
from renate.evaluation.metrics.utils import create_metrics
from renate.memory import DataBuffer, InfiniteBuffer, ReservoirBuffer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.utils.optimizer import create_optimizer, create_scheduler
from renate.utils.pytorch import get_generator


class Learner(LightningModule, abc.ABC):
    """Base class for Learners, which encapsulate the core CL methodologies.

    The `Learner` is a `LightningModule`, but provides additional hook functions
    called by `ModelUpdater`. These hooks are:

    - `Learner.on_model_update_start`, which is called in the beginning of a
       model update. We expect this to return train and (optionally) validation
       data loader(s).
    - `Learner.on_model_update_end`, which is called in the end of a model update.

    This base class implements a basic training loop without any mechanism to
    counteract forgetting.

    Args:
        model: The model to be trained.
        optimizer: Optimizer used for training. Options: `Adam` or `SGD`.
        learning_rate: Initial learning rate used for training.
        learning_rate_scheduler: Learning rate scheduler used for training.
        learning_rate_scheduler_gamma: Learning rate scheduler gamma.
        learning_rate_scheduler_step_size: Learning rate scheduler step size.
        momentum: Momentum term (only relevant for optimizer `SGD`).
        weight_decay: L2 regularization applied to all model weights.
        batch_size: Training batch size.
        train_transform: The transformation applied during training.
        train_target_transform: The target transformation applied during testing.
        test_transform: The transformation at test time.
        test_target_transform: The target transformation at test time.
        logged_metrics: Metrics logged additional to the default ones.
        seed: See :func:`renate.models.utils.get_generator`.
    """

    def __init__(
        self,
        model: RenateModule,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        seed: int = defaults.SEED,
    ) -> None:
        super().__init__()
        self._model = model
        self._optimizer = optimizer
        self._learning_rate = learning_rate
        self._learning_rate_scheduler = learning_rate_scheduler
        self._learning_rate_scheduler_gamma = learning_rate_scheduler_gamma
        self._learning_rate_scheduler_step_size = learning_rate_scheduler_step_size
        self._momentum = momentum
        self._weight_decay = weight_decay
        self._batch_size = batch_size
        self._train_transform = train_transform
        self._train_target_transform = train_target_transform
        self._test_transform = test_transform
        self._test_target_transform = test_target_transform
        self._seed = seed
        self._task_id: str = defaults.TASK_ID

        self._val_memory_buffer: DataBuffer = InfiniteBuffer()
        self._create_metrics_collections(logged_metrics)
        self._post_init()

    def _post_init(self) -> None:
        self._rng = get_generator(self._seed)

    def _create_metrics_collections(
        self, logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None
    ) -> None:
        """Creates all logged metrics."""
        metrics = create_metrics(task="classification", additional_metrics=logged_metrics)
        train_metrics = metrics.clone(prefix="train_")
        val_metrics = metrics.clone(prefix="val_")

        train_losses = nn.ModuleDict(
            {
                "base_loss": torchmetrics.MeanMetric(),
                "loss": torchmetrics.MeanMetric(),
            }
        )
        val_losses = nn.ModuleDict({"loss": torchmetrics.MeanMetric()})

        self._metric_collections = nn.ModuleDict(
            {
                "train_metrics": train_metrics,
                "val_metrics": val_metrics,
            }
        )
        self._loss_collections = nn.ModuleDict(
            {
                "train_losses": train_losses,
                "val_losses": val_losses,
            }
        )

    def state_dict(self, **kwargs) -> Dict[str, Any]:
        """Returns the state of the learner."""
        return {
            "learner_class_name": self.__class__.__name__,
            "optimizer": self._optimizer,
            "learning_rate": self._learning_rate,
            "learning_rate_scheduler": self._learning_rate_scheduler,
            "learning_rate_scheduler_gamma": self._learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": self._learning_rate_scheduler_step_size,
            "momentum": self._momentum,
            "weight_decay": self._weight_decay,
            "batch_size": self._batch_size,
            "seed": self._seed,
            "task_id": self._task_id,
            "val_memory_buffer": self._val_memory_buffer.state_dict(),
        }

    def load_state_dict(self, model: RenateModule, state_dict: Dict[str, Any], **kwargs) -> None:
        """Restores the state of the learner.

        Even though this is a LightningModule, no modules are stored.

        Args:
            model: The model to be trained.
            state_dict: Dictionary containing the state.
        """
        if self.__class__.__name__ != state_dict["learner_class_name"]:
            raise RuntimeError(
                f"Learner of class {self.__class__} was used to load a state dict created by class "
                f"{state_dict['learner_class_name']}."
            )
        super().__init__()
        self._model = model
        self._optimizer = state_dict["optimizer"]
        self._learning_rate = state_dict["learning_rate"]
        self._learning_rate_scheduler = state_dict["learning_rate_scheduler"]
        self._learning_rate_scheduler_gamma = state_dict["learning_rate_scheduler_gamma"]
        self._learning_rate_scheduler_step_size = state_dict["learning_rate_scheduler_step_size"]
        self._momentum = state_dict["momentum"]
        self._weight_decay = state_dict["weight_decay"]
        self._batch_size = state_dict["batch_size"]
        self._seed = state_dict["seed"]
        self._task_id = state_dict["task_id"]
        if not hasattr(self, "_val_memory_buffer"):
            self._val_memory_buffer = InfiniteBuffer()
        self._val_memory_buffer.load_state_dict(state_dict["val_memory_buffer"])
        self._post_init()

    def save(self, output_state_dir: str) -> None:
        val_buffer_dir = os.path.join(output_state_dir, "val_memory_buffer")
        os.makedirs(val_buffer_dir, exist_ok=True)
        self._val_memory_buffer.save(val_buffer_dir)

    def load(self, input_state_dir: str) -> None:
        self._val_memory_buffer.load(os.path.join(input_state_dir, "val_memory_buffer"))

    def set_transforms(
        self,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
    ) -> None:
        """Update the transformations applied to the data."""
        self._train_transform = train_transform
        self._train_target_transform = train_target_transform
        self._test_transform = test_transform
        self._test_target_transform = test_target_transform

    def set_logged_metrics(
        self, logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None
    ) -> None:
        """Sets the additional metrics logged during training and evaluation."""
        self._create_metrics_collections(logged_metrics)

    def is_logged_metric(self, metric_name: str) -> bool:
        """Returns `True` if there is a metric with name `metric_name`."""
        if metric_name is None:
            return True
        logged_metrics = list()
        for prefix in ["train", "val"]:
            for collection, collection_name in zip(
                [self._metric_collections, self._loss_collections], ["metrics", "losses"]
            ):
                collection_key = f"{prefix}_{collection_name}"
                if collection_key in collection:
                    logged_metrics += [
                        f"{prefix}_{logged_metric_name}"
                        for logged_metric_name in collection[collection_key]
                    ]
        return metric_name in logged_metrics

    def update_hyperparameters(self, args: Dict[str, Any]) -> None:
        """Update the hyperparameters of the learner."""
        if "optimizer" in args:
            self._optimizer = args["optimizer"]
        if "learning_rate" in args:
            self._learning_rate = args["learning_rate"]
        if "learning_rate_scheduler" in args:
            self._learning_rate_scheduler = args["learning_rate_scheduler"]
        if "learning_rate_scheduler_gamma" in args:
            self._learning_rate_scheduler_gamma = args["learning_rate_scheduler_gamma"]
        if "learning_rate_scheduler_step_size" in args:
            self._learning_rate_scheduler_step_size = args["learning_rate_scheduler_step_size"]
        if "momentum" in args:
            self._momentum = args["momentum"]
        if "weight_decay" in args:
            self._weight_decay = args["weight_decay"]
        if "batch_size" in args:
            self._batch_size = args["batch_size"]

    def on_model_update_start(
        self, train_dataset: Dataset, val_dataset: Dataset, task_id: Optional[str] = None
    ) -> None:
        self._train_dataset = train_dataset
        self._val_dataset = val_dataset
        self.val_enabled = val_dataset is not None and len(val_dataset)
        self._task_id = task_id
        self._model.add_task_params(task_id=self._task_id)

    def train_dataloader(self) -> DataLoader:
        """Returns the dataloader for training the model."""
        train_dataset = _TransformedDataset(
            self._train_dataset,
            transform=self._train_transform,
            target_transform=self._train_target_transform,
        )
        return DataLoader(
            train_dataset,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
        )

    def val_dataloader(self) -> DataLoader:
        if self._val_dataset is not None:
            val_dataset = _TransformedDataset(
                self._val_dataset,
                transform=self._test_transform,
                target_transform=self._test_target_transform,
            )
            self._val_memory_buffer.update(val_dataset)

        if len(self._val_memory_buffer):
            return DataLoader(
                self._val_memory_buffer,
                batch_size=self._batch_size,
                shuffle=False,
                generator=self._rng,
                pin_memory=True,
            )
        else:
            return None

    def on_model_update_end(self) -> None:
        """Called right before a model update terminates."""
        pass

    def forward(self, inputs: NestedTensors, task_id: Optional[str] = None) -> torch.Tensor:
        """Forward pass of the model."""
        if task_id is None:
            task_id = self._task_id
        return self._model(inputs, task_id=task_id)

    def training_step(
        self, batch: Tuple[NestedTensors, torch.Tensor], batch_idx: int
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        inputs, targets = batch
        outputs = self(inputs)
        intermediate_representation = self._model.get_intermediate_representation()
        self._model.reset_intermediate_representation_cache()
        loss = self._model.loss_fn(outputs, targets)
        self._update_metrics(outputs, targets, "train")
        self._loss_collections["train_losses"]["base_loss"](loss)
        return {
            "loss": loss,
            "outputs": outputs,
            "intermediate_representation": intermediate_representation,
        }

    def training_step_end(self, step_output: STEP_OUTPUT) -> STEP_OUTPUT:
        """PyTorch Lightning function to perform after the training step."""
        super().training_step_end(step_output)
        self._loss_collections["train_losses"]["loss"](step_output["loss"])
        return step_output

    def training_epoch_end(self, outputs: List[Union[Tensor, Dict[str, Any]]]) -> None:
        """PyTorch Lightning function to run at the end of training epoch."""
        super().training_epoch_end(outputs)
        if not self.val_enabled:
            self._log_metrics()

    def validation_step(self, batch: Tuple[NestedTensors, torch.Tensor], batch_idx: int) -> None:
        """PyTorch Lightning function to estimate validation metrics."""
        (inputs, targets), _ = batch
        outputs = self(inputs)
        loss = self._model.loss_fn(outputs, targets)
        self._update_metrics(outputs, targets, "val")
        self._loss_collections["val_losses"]["loss"](loss)

    def validation_epoch_end(self, outputs: List[Union[Tensor, Dict[str, Any]]]) -> None:
        """PyTorch Lightning function to run at the end of validation epoch."""
        super().validation_epoch_end(outputs)
        self._log_metrics()

    def configure_optimizers(
        self,
    ) -> Tuple[List[torch.optim.Optimizer], List[torch.optim.lr_scheduler._LRScheduler]]:
        """PyTorch Lightning function to create an optimizer."""
        optimizer = create_optimizer(
            params=self._model.get_params(self._task_id),
            optimizer=self._optimizer,
            lr=self._learning_rate,
            momentum=self._momentum,
            weight_decay=self._weight_decay,
        )
        scheduler = create_scheduler(
            scheduler=self._learning_rate_scheduler,
            optimizer=optimizer,
            gamma=self._learning_rate_scheduler_gamma,
            step_size=self._learning_rate_scheduler_step_size,
        )
        return [optimizer], [scheduler]

    def _update_metrics(
        self,
        outputs: torch.Tensor,
        y: torch.Tensor,
        prefix: Literal["train", "val"],
    ) -> None:
        """Shared logic for updating metrics."""
        self._metric_collections[f"{prefix}_metrics"](outputs, y)

    def _log_metrics(
        self,
    ) -> None:
        """Shared logic for logging metrics, including the loss."""
        if self.trainer.sanity_checking:
            return
        prefixes = ["train", "val"] if self.val_enabled else ["train"]
        for prefix in prefixes:
            self.log_dict(
                self._metric_collections[f"{prefix}_metrics"].compute(),
                on_step=False,
                on_epoch=True,
                logger=True,
            )
            self._metric_collections[f"{prefix}_metrics"].reset()

            for loss_name, loss in self._loss_collections[f"{prefix}_losses"].items():
                self.log(
                    f"{prefix}_{loss_name}",
                    loss.compute(),
                    on_step=False,
                    on_epoch=True,
                    logger=True,
                )
                loss.reset()


class ReplayLearner(Learner, abc.ABC):
    """Base class for Learners which use a buffer to store data and reuse it in future updates.

    Args:
        memory_size: The maximum size of the memory.
        memory_batch_size: Size of batches sampled from the memory. The memory batch will be
            appended to the batch sampled from the current dataset, leading to an effective batch
            size of `memory_batch_size + batch_size`.
        buffer_transform: The transformation to be applied to the memory buffer data samples.
        buffer_target_transform: The target transformation to be applied to the memory buffer target
            samples.
        seed: See :func:`renate.models.utils.get_generator`.
    """

    def __init__(
        self,
        memory_size: int,
        memory_batch_size: int = defaults.BATCH_SIZE,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        seed: int = defaults.SEED,
        **kwargs,
    ) -> None:
        super().__init__(seed=seed, **kwargs)
        self._memory_batch_size = min(memory_size, memory_batch_size)
        self._memory_buffer = ReservoirBuffer(
            max_size=memory_size,
            seed=seed,
            transform=buffer_transform,
            target_transform=buffer_target_transform,
        )

    def state_dict(self, **kwargs) -> Dict[str, Any]:
        """Returns the state of the learner."""
        state_dict = super().state_dict(**kwargs)
        state_dict.update(
            {
                "memory_batch_size": self._memory_batch_size,
                "memory_buffer": self._memory_buffer.state_dict(),
            }
        )
        return state_dict

    def load_state_dict(self, model: RenateModule, state_dict: Dict[str, Any], **kwargs) -> None:
        """Restores the state of the learner."""
        super().load_state_dict(model, state_dict, **kwargs)
        self._memory_batch_size = state_dict["memory_batch_size"]
        if not hasattr(self, "_memory_buffer"):
            self._memory_buffer = ReservoirBuffer()
        self._memory_buffer.load_state_dict(state_dict["memory_buffer"])

    def save(self, output_state_dir: str) -> None:
        super().save(output_state_dir)
        buffer_dir = os.path.join(output_state_dir, "memory_buffer")
        os.makedirs(buffer_dir, exist_ok=True)
        self._memory_buffer.save(buffer_dir)

    def load(self, input_state_dir: str) -> None:
        super().load(input_state_dir)
        self._memory_buffer.load(os.path.join(input_state_dir, "memory_buffer"))

    def set_transforms(
        self,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
    ) -> None:
        """Update the transformations applied to the data."""
        super().set_transforms(
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
        )
        self._memory_buffer.set_transforms(buffer_transform, buffer_target_transform)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Callable, Dict, Optional, Tuple

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.memory import GreedyClassBalancingBuffer
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner import Learner, ReplayLearner
from renate.updaters.model_updater import SingleTrainingLoopUpdater
from renate.utils.pytorch import reinitialize_model_parameters


class GDumbLearner(ReplayLearner):
    """A Learner that implements the GDumb strategy.

    Prabhu, Ameya, Philip HS Torr, and Puneet K. Dokania. "GDumb: A simple
    approach that questions our progress in continual learning." ECCV, 2020.

    It maintains a memory of  previously observed data points and does
    the training after updating the buffer. Note that, the model is
    reinitialized before training on the buffer.

    Args:
        memory_size: The maximum size of the memory.
        buffer_transform: The transform to be applied to the data points in the memory.
        buffer_target_transform: The transform to be applied to the targets in the memory.
        seed: A random seed.
    """

    def __init__(
        self,
        memory_size: int,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        seed: int = defaults.SEED,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            memory_size=memory_size,
            seed=seed,
            **kwargs,
        )
        self._memory_buffer = GreedyClassBalancingBuffer(
            max_size=memory_size,
            seed=seed,
            transform=buffer_transform,
            target_transform=buffer_target_transform,
        )

    def load_state_dict(self, model: RenateModule, state_dict: Dict[str, Any], **kwargs) -> None:
        """Restores the state of the learner."""
        if not hasattr(self, "_memory_buffer"):
            self._memory_buffer = GreedyClassBalancingBuffer()
        Learner.load_state_dict(self, model, state_dict, **kwargs)
        self._memory_batch_size = state_dict["memory_batch_size"]
        self._memory_buffer.load_state_dict(state_dict["memory_buffer"])

    def on_model_update_start(
        self, train_dataset: Dataset, val_dataset: Dataset, task_id: Optional[str] = None
    ) -> None:
        """Called before a model update starts."""
        super().on_model_update_start(train_dataset, val_dataset, task_id)
        self._memory_buffer.update(train_dataset)
        reinitialize_model_parameters(self._model)

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self._memory_buffer,
            batch_size=self._batch_size,
            shuffle=True,
            generator=self._rng,
            pin_memory=True,
        )

    def training_step(
        self,
        batch: Tuple[Tuple[NestedTensors, torch.Tensor], Dict[str, torch.Tensor]],
        batch_idx: int,
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        batch, _ = batch
        return super().training_step(batch=batch, batch_idx=batch_idx)


class GDumbModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        memory_size: int,
        memory_batch_size: int = defaults.BATCH_SIZE,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "memory_batch_size": memory_batch_size,
            "optimizer": optimizer,
            "learning_rate": learning_rate,
            "learning_rate_scheduler": learning_rate_scheduler,
            "learning_rate_scheduler_gamma": learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": learning_rate_scheduler_step_size,
            "momentum": momentum,
            "weight_decay": weight_decay,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            learner_class=GDumbLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            deterministic_trainer=deterministic_trainer,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import torch
import torchvision
from torch.utils.data import Dataset
from torchvision import transforms

from renate import defaults
from renate.data import ImageDataset
from renate.data.data_module import RenateDataModule
from renate.utils.file import download_and_unzip_file, download_folder_from_s3


class TinyImageNetDataModule(RenateDataModule):
    """Datamodule that process TinyImageNet dataset.

    Source: http://cs231n.stanford.edu/

    The TinyImageNet dataset is a subset of the ImageNet dataset. It contains 200 classes, each with
    500 training images, 50 validation images with labels. There are also 50 unlabeled test images
    per class, which we are not using here. We use the validation split as the test set.

    Args:
        data_path: Path to the directory where the dataset should be stored.
        src_bucket: Name of the bucket where the dataset is stored.
        src_object_name: Name of the object in the bucket where the dataset is stored.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed to be used for splitting the dataset.
    """

    md5s = {"tiny-imagenet-200.zip": "90528d7ca1a48142e341f4ef8d21d0de"}

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ) -> None:
        super(TinyImageNetDataModule, self).__init__(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = "tiny_imagenet"

    def prepare_data(self) -> None:
        """Download the TinyImageNet dataset."""
        if not self._verify_file("tiny-imagenet-200.zip"):
            download_and_unzip_file(
                self._dataset_name,
                self._data_path,
                self._src_bucket,
                self._src_object_name,
                "http://cs231n.stanford.edu/",
                "tiny-imagenet-200.zip",
            )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        X, y = self._preprocess_tiny_imagenet(train=True)
        train_data = ImageDataset(X, y, transform=transforms.ToTensor())
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        X, y = self._preprocess_tiny_imagenet(train=False)
        self._test_data = ImageDataset(X, y, transform=transforms.ToTensor())

    def _preprocess_tiny_imagenet(self, train: bool) -> Tuple[List[str], List[int]]:
        """A helper function to preprocess the TinyImageNet dataset."""
        if train:
            data_path = os.path.join(self._data_path, self._dataset_name, "tiny-imagenet-200")
        else:
            data_path = os.path.join(self._data_path, self._dataset_name, "tiny-imagenet-200")
        with open(os.path.join(data_path, "wnids.txt"), "r") as f:
            label_encoding = {line.strip(): i for i, line in enumerate(f.readlines())}
        X = []
        y = []
        if train:
            for label in label_encoding.keys():
                class_path = os.path.join(data_path, "train", label)
                for file_name in os.listdir(os.path.join(class_path, "images")):
                    X.append(os.path.join(class_path, "images", file_name))
                    y.append(label_encoding[label])
        else:
            val_annotations_file = os.path.join(data_path, "val", "val_annotations.txt")
            with open(val_annotations_file, "r") as f:
                for line in f:
                    file_name, label = line.split("\t")[:2]
                    X.append(os.path.join(data_path, "val", "images", file_name))
                    y.append(label_encoding[label])
        return X, y


class TorchVisionDataModule(RenateDataModule):
    """Data module wrapping torchvision datasets.

    Args:
        data_path: the path to the folder containing the dataset files.
        src_bucket: the name of the s3 bucket. If not provided, downloads the data from original
            source.
        src_object_name: the folder path in the s3 bucket.
        dataset_name: Name of the torchvision dataset.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    dataset_dict = {
        "CIFAR10": (torchvision.datasets.CIFAR10, "cifar-10-batches-py"),
        "CIFAR100": (torchvision.datasets.CIFAR100, "cifar-100-python"),
        "FashionMNIST": (torchvision.datasets.FashionMNIST, "FashionMNIST"),
        "MNIST": (torchvision.datasets.MNIST, "MNIST"),
    }
    dataset_stats = {
        "CIFAR10": {
            "mean": (0.49139967861519607, 0.48215840839460783, 0.44653091444546567),
            "std": (0.24703223246174102, 0.24348512800151828, 0.26158784172803257),
        },
        "CIFAR100": {
            "mean": (0.5070751592371323, 0.48654887331495095, 0.4409178433670343),
            "std": (0.26733428587941854, 0.25643846292120615, 0.2761504713263903),
        },
        "FashionMNIST": {"mean": 0.2860405969887955, "std": 0.3530242445149223},
        "MNIST": {"mean": 0.1306604762738429, "std": 0.30810780385646264},
    }

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        dataset_name: str = "MNIST",
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super(TorchVisionDataModule, self).__init__(
            data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = dataset_name
        if self._dataset_name not in TorchVisionDataModule.dataset_dict:
            raise ValueError(f"Dataset {self._dataset_name} currently not supported.")

    def prepare_data(self) -> None:
        """Download data.

        If s3 bucket is given, the data is downloaded from s3, otherwise from the original source.
        """
        cls, dataset_pathname = TorchVisionDataModule.dataset_dict[self._dataset_name]
        if self._src_bucket is None:
            cls(self._data_path, train=True, download=True)
            cls(self._data_path, train=False, download=True)
        else:
            download_folder_from_s3(
                src_bucket=self._src_bucket,
                src_object_name=self._src_object_name,
                dst_dir=os.path.join(self._data_path, dataset_pathname),
            )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        cls, _ = TorchVisionDataModule.dataset_dict[self._dataset_name]
        train_data = cls(
            self._data_path,
            train=True,
            transform=transforms.ToTensor(),
            target_transform=transforms.Lambda(lambda x: torch.tensor(x, dtype=torch.long)),
        )
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self._test_data = cls(
            self._data_path,
            train=False,
            transform=transforms.ToTensor(),
            target_transform=transforms.Lambda(lambda x: torch.tensor(x, dtype=torch.long)),
        )


class CLEARDataModule(RenateDataModule):
    """Datamodule that process CLEAR datasets: CLEAR10 and CLEAR100.

    Source: https://clear-benchmark.github.io/.

    Args:
        data_path: the path to the folder containing the dataset files.
        src_bucket: the name of the s3 bucket. If not provided, downloads the data from original
            source.
        src_object_name: the folder path in the s3 bucket.
        dataset_name: CLEAR dataset name, options are clear10 and clear100.
        chunk_id: Used to define the CLEAR dataset splits. There are 10 splits in total with ids
            from 0 to 9.
        val_size: Fraction of the training data to be used for validation.
        seed: Seed used to fix random number generation.
    """

    md5s = {
        "clear10-train-image-only.zip": "5171f720810d60b471c308dee595d430",
        "clear100-train-image-only.zip": "ea85cdba9efcb3abf77eaab5554052c8",
        "clear10-test.zip": "bf9a85bfb78fe742c7ed32648c9a3275",
        "clear100-test.zip": "e160815fb5fd4bc71dacd339ff41e6a9",
    }

    def __init__(
        self,
        data_path: Union[Path, str],
        src_bucket: Optional[str] = None,
        src_object_name: Optional[str] = None,
        dataset_name: str = "CLEAR10",
        chunk_id: int = 0,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super(CLEARDataModule, self).__init__(
            data_path=data_path,
            src_bucket=src_bucket,
            src_object_name=src_object_name,
            val_size=val_size,
            seed=seed,
        )
        self._dataset_name = dataset_name.lower()
        assert self._dataset_name in ["clear10", "clear100"]
        self._verify_chunk_id(chunk_id)
        self._chunk_id = chunk_id

    def _verify_chunk_id(self, chunk_id: int) -> None:
        """Verify that the chunk_id is valid."""
        assert 0 <= chunk_id <= 9

    def prepare_data(self) -> None:
        """Download CLEAR dataset with given dataset_name (clear10/clear100)."""
        for file_name in [
            f"{self._dataset_name}-train-image-only.zip",
            f"{self._dataset_name}-test.zip",
        ]:
            if not self._verify_file(file_name):
                download_and_unzip_file(
                    self._dataset_name,
                    self._data_path,
                    self._src_bucket,
                    self._src_object_name,
                    "https://clear-challenge.s3.us-east-2.amazonaws.com/",
                    file_name,
                )

    def setup(self) -> None:
        """Set up train, test and val datasets."""
        X, y = self._get_filepaths_and_labels(train=True, chunk_id=self._chunk_id)
        train_data = ImageDataset(X, y, transform=transforms.ToTensor())
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self._test_data = []
        for i in range(10):
            X, y = self._get_filepaths_and_labels(train=False, chunk_id=i)
            self._test_data.append(ImageDataset(X, y, transform=transforms.ToTensor()))

    # TODO: This is a work-around to make CLEAR available as a data module (single test dataset)
    # as well as a scenario (all test datasets) via BenchmarkScenario. Clean up!
    def test_data(self) -> Dataset:
        return self._test_data[self._chunk_id]

    def _get_filepaths_and_labels(self, train: bool, chunk_id: int) -> Tuple[List[str], List[int]]:
        """Extracts all the filepaths and labels for a given chunk id and split."""
        data = []
        labels = []
        path = os.path.join(self._data_path, self._dataset_name)
        path = os.path.join(path, "train_image_only" if train else "test")

        # Load the class names and create a class mapping. The class names are in `class_names.txt`
        label_encoding = {}
        with open(os.path.join(path, "class_names.txt"), "r") as f:
            class_names = [line.strip() for line in f.readlines() if line.strip() != "BACKGROUND"]
            label_encoding = {name: cnt for cnt, name in enumerate(class_names)}

        path = os.path.join(path, "labeled_images", str(chunk_id + 1))

        # Go through all the subfolders in the path folder and search for all .jpg images
        for root, _, files in os.walk(path):
            for file in files:
                if file.endswith(".jpg"):
                    folder = root.split("/")[-1]
                    if folder == "BACKGROUND":
                        continue
                    data.append(os.path.join(root, file))
                    labels.append(label_encoding[folder])

        return data, labels



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Callable, Dict, Optional, Tuple

import torch
import torchmetrics
from pytorch_lightning.loggers.logger import Logger
from pytorch_lightning.trainer.supporters import CombinedLoader
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.utils.data import DataLoader, Dataset

from renate import defaults
from renate.models import RenateModule
from renate.types import NestedTensors
from renate.updaters.learner import ReplayLearner
from renate.updaters.model_updater import SingleTrainingLoopUpdater


class OfflineExperienceReplayLearner(ReplayLearner):
    """Experience Replay in the offline version.

    The model will be trained on weighted mixture of losses computed on the new data and a replay
    buffer. In contrast to the online version, the buffer will only be updated after training has
    terminated.

    Args:
        memory_size: The maximum size of the memory.
        memory_batch_size: Size of batches sampled from the memory. The memory batch will be
            appended to the batch sampled from the current dataset, leading to an effective batch
            size of `memory_batch_size + batch_size`.
        loss_weight_new_data: The training loss will be a convex combination of the loss on the new
            data and the loss on the memory data. If a float (needs to be in [0, 1]) is given here,
            it will be used as the weight for the new data. If `None`, the weight will be set
            dynamically to `N_t / sum([N_1, ..., N_t])`, where `N_i` denotes the size of task/chunk
            `i` and the current task is `t`.
        buffer_transform: The transformation to be applied to the memory buffer data samples.
        buffer_target_transform: The target transformation to be applied to the memory buffer target
            samples.
    """

    def __init__(
        self,
        loss_weight_new_data: Optional[float] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if loss_weight_new_data is not None and not (0.0 <= loss_weight_new_data <= 1.0):
            raise ValueError(
                "Value of loss_weight_new_data needs to be between 0 and 1,"
                f"got {loss_weight_new_data}."
            )
        self._loss_weight_new_data = loss_weight_new_data
        self._num_points_previous_tasks: int = 0

    def _create_metrics_collections(
        self, logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None
    ) -> None:
        super()._create_metrics_collections(logged_metrics)
        self._loss_collections["train_losses"]["memory_loss"] = torchmetrics.MeanMetric()

    def on_model_update_start(
        self, train_dataset: Dataset, val_dataset: Dataset, task_id: Optional[str] = None
    ) -> None:
        """Called before a model update starts."""
        super().on_model_update_start(train_dataset, val_dataset, task_id)
        self._num_points_current_task = len(train_dataset)

    def train_dataloader(self) -> DataLoader:
        train_loader = super().train_dataloader()
        loaders = {"current_task": train_loader}
        if len(self._memory_buffer) > self._memory_batch_size:
            loaders["memory"] = DataLoader(
                dataset=self._memory_buffer,
                batch_size=self._memory_batch_size,
                drop_last=True,
                shuffle=True,
                generator=self._rng,
                pin_memory=True,
            )
        return CombinedLoader(loaders, mode="max_size_cycle")

    def on_model_update_end(self) -> None:
        """Called right before a model update terminates."""
        self._memory_buffer.update(self._train_dataset)
        self._num_points_previous_tasks += self._num_points_current_task
        self._num_points_current_task = -1

    def training_step(
        self, batch: Dict[str, Tuple[NestedTensors, torch.Tensor]], batch_idx: int
    ) -> STEP_OUTPUT:
        """PyTorch Lightning function to return the training loss."""
        if self._loss_weight_new_data is None:
            alpha = self._num_points_current_task / (
                self._num_points_current_task + self._num_points_previous_tasks
            )
        else:
            alpha = self._loss_weight_new_data
        inputs, targets = batch["current_task"]
        outputs = self(inputs)
        loss = self._model.loss_fn(outputs, targets)
        self._loss_collections["train_losses"]["base_loss"](loss)
        self._update_metrics(outputs, targets, "train")
        if "memory" in batch:
            (inputs_mem, targets_mem), _ = batch["memory"]
            outputs_mem = self(inputs_mem)
            loss_mem = self._model.loss_fn(outputs_mem, targets_mem)
            self._loss_collections["train_losses"]["memory_loss"](loss_mem)
            loss = alpha * loss + (1.0 - alpha) * loss_mem
        return {"loss": loss}

    def state_dict(self, **kwargs) -> Dict[str, Any]:
        """Returns the state of the learner."""
        state_dict = super().state_dict(**kwargs)
        state_dict["loss_weight_new_data"] = self._loss_weight_new_data
        state_dict["num_points_previous_tasks"] = self._num_points_previous_tasks
        return state_dict

    def load_state_dict(self, model: RenateModule, state_dict: Dict[str, Any], **kwargs) -> None:
        """Restores the state of the learner."""
        super().load_state_dict(model, state_dict, **kwargs)
        self._loss_weight_new_data = state_dict["loss_weight_new_data"]
        self._num_points_previous_tasks = state_dict["num_points_previous_tasks"]


class OfflineExperienceReplayModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        memory_size: int,
        memory_batch_size: int = defaults.BATCH_SIZE,
        loss_weight_new_data: Optional[float] = None,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "memory_batch_size": memory_batch_size,
            "loss_weight_new_data": loss_weight_new_data,
            "optimizer": optimizer,
            "learning_rate": learning_rate,
            "learning_rate_scheduler": learning_rate_scheduler,
            "learning_rate_scheduler_gamma": learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": learning_rate_scheduler_step_size,
            "momentum": momentum,
            "weight_decay": weight_decay,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            learner_class=OfflineExperienceReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            deterministic_trainer=deterministic_trainer,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Callable, Optional

import torch
from torch.utils.data import Dataset

from renate import defaults
from renate.data.data_module import RenateDataModule
from renate.utils.pytorch import get_generator


class DummyDataset(Dataset):
    def __init__(self, data, targets, transform=None):
        assert len(data) == len(targets)
        self.data = data
        self.targets = targets
        self.transform = transform

    def __getitem__(self, index):
        x = self.data[index]
        y = self.targets[index]
        if self.transform is not None:
            x = self.transform(x)
        return x, y

    def __len__(self):
        return len(self.data)


class DummyTorchVisionDataModule(RenateDataModule):
    """
    A simple data module similar to `TorchVisionDataModule` with 100 training instances and 100 test
    instances with shape (1, 5, 5) and 5 classes.
    """

    def __init__(
        self,
        transform: Optional[Callable] = None,
        val_size: float = defaults.VALIDATION_SIZE,
        seed: int = defaults.SEED,
    ):
        super().__init__(
            data_path="",
            src_bucket="",
            src_object_name="",
            val_size=val_size,
            seed=seed,
        )
        self._transform = transform
        self.input_shape = (1, 5, 5)
        self.output_shape = (1,)
        self.output_classes = 5

    def prepare_data(self) -> None:
        pass

    def _get_random_data(self, offset: int = 0):
        rng = get_generator(self._seed + offset)
        X = torch.randint(0, 42, (100, *self.input_shape), generator=rng, dtype=torch.float32)
        y = torch.ones(100, dtype=torch.long)
        for i in range(5):
            y[i * 20 : (i + 1) * 20] = i
        return X, y

    def setup(self):
        self.X_train, self.y_train = self._get_random_data()
        train_data = DummyDataset(self.X_train, self.y_train, self._transform)
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self.X_test, self.y_test = self._get_random_data()
        self._test_data = DummyDataset(self.X_test, self.y_test, self._transform)


class DummyTorchVisionDataModuleWithChunks(DummyTorchVisionDataModule):
    """This dataset has multiple chunks that can be queried through chunk ID."""

    def __init__(
        self,
        transform=None,
        val_size=defaults.VALIDATION_SIZE,
        seed=defaults.SEED,
        chunk_id=0,
        num_chunks=1,
    ):
        super().__init__(transform=transform, val_size=val_size, seed=seed)
        self.chunk_id = chunk_id
        self.num_chunks = num_chunks

    def setup(self):
        super().setup()
        train_data = DummyDataset(
            torch.split(self.X_train, 100 // self.num_chunks)[self.chunk_id],
            torch.split(self.y_train, 100 // self.num_chunks)[self.chunk_id],
            self._transform,
        )
        self._train_data, self._val_data = self._split_train_val_data(train_data)
        self._test_data = []
        for i in range(self.num_chunks):
            self._test_data.append(
                DummyDataset(
                    torch.split(self.X_test, 100 // self.num_chunks)[i],
                    torch.split(self.y_test, 100 // self.num_chunks)[i],
                    self._transform,
                )
            )

    # TODO: This is a work-around to make CLEAR available as a data module (single test dataset)
    # as well as a scenario (all test datasets) via BenchmarkScenario. Clean up!
    def test_data(self) -> Dataset:
        return self._test_data[self.chunk_id]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from functools import partial
from typing import Any, Callable, List, Optional

import torch.nn as nn
from torchvision.models.vision_transformer import ConvStemConfig, WeightsEnum
from torchvision.models.vision_transformer import VisionTransformer as _VisionTransformer

from renate.benchmark.models.base import RenateBenchmarkingModule
from renate.models.prediction_strategies import PredictionStrategy


class VisionTransformer(RenateBenchmarkingModule):
    """Vision Transformer base model.

    TODO: Fix citation
    Dosovitskiy, Alexey, et al. "An image is worth 16x16 words: Transformers for image recognition
    at scale."
    arXiv preprint arXiv:2010.11929 (2020).

    Args:
        image_size: Size of the input image.
        patch_size: Size of the patches.
        num_layers: Number of Encoder layers.
        num_heads: Number of Attention heads.
        hidden_dim: Size of the Encoder's hidden state.
        mlp_dim: Size of the intermediate Multi-layer Perceptron in the Encoder.
        dropout: Dropout probability.
        attention_dropout: Dropout probability for the attention in the Multi-head Attention layer.
        num_outputs: Size of the output.
        representation_size: If specified, the model will return a linear projection of the last
            hidden state.
        norm_layer: Normalization layer.
        conv_stem_configs: List of ConvStemConfig. Each ConvStemConfig corresponds to a
            convolutional stem.
        loss: Loss function.
        prediction_strategy: Continual learning strategies may alter the prediction at train or test
            time.
        add_icarl_class_means: If ``True``, additional parameters used only by the
            ``ICaRLModelUpdater`` are added. Only required when using that updater.
    """

    def __init__(
        self,
        image_size: int = 32,
        patch_size: int = 4,
        num_layers: int = 12,
        num_heads: int = 12,
        hidden_dim: int = 768,
        mlp_dim: int = 3072,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        num_outputs: int = 10,
        representation_size: Optional[int] = None,
        norm_layer: Callable[..., nn.Module] = partial(nn.LayerNorm, eps=1e-6),
        conv_stem_configs: Optional[List[ConvStemConfig]] = None,
        weights: Optional[WeightsEnum] = None,
        loss: nn.Module = nn.CrossEntropyLoss(),
        prediction_strategy: Optional[PredictionStrategy] = None,
        add_icarl_class_means: bool = True,
    ) -> None:
        model = _VisionTransformer(
            image_size=image_size,
            patch_size=patch_size,
            num_layers=num_layers,
            num_heads=num_heads,
            hidden_dim=hidden_dim,
            mlp_dim=mlp_dim,
            dropout=dropout,
            attention_dropout=attention_dropout,
            num_classes=num_outputs,
            representation_size=representation_size,
            norm_layer=norm_layer,
            conv_stem_configs=conv_stem_configs,
        )
        super().__init__(
            embedding_size=model.heads.head.in_features,
            num_outputs=num_outputs,
            constructor_arguments={
                "image_size": image_size,
                "patch_size": patch_size,
                "num_layers": num_layers,
                "num_heads": num_heads,
                "hidden_dim": hidden_dim,
                "mlp_dim": mlp_dim,
                "dropout": dropout,
                "attention_dropout": attention_dropout,
                "representation_size": representation_size,
                "norm_layer": norm_layer,
                "conv_stem_configs": conv_stem_configs,
            },
            loss_fn=loss,
            prediction_strategy=prediction_strategy,
            add_icarl_class_means=add_icarl_class_means,
        )
        self._backbone = model
        if weights:
            self._backbone.load_state_dict(weights.get_state_dict())
        self._backbone.heads.head = nn.Identity()


class VisionTransformerCIFAR(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            image_size=32,
            patch_size=4,
            num_layers=3,
            num_heads=3,
            hidden_dim=192,
            mlp_dim=768,
            **kwargs,
        )


class VisionTransformerB16(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            image_size=224,
            patch_size=16,
            num_layers=12,
            num_heads=12,
            hidden_dim=768,
            mlp_dim=3072,
            **kwargs,
        )


class VisionTransformerB32(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            image_size=224,
            patch_size=32,
            num_layers=12,
            num_heads=12,
            hidden_dim=768,
            mlp_dim=3072,
            **kwargs,
        )


class VisionTransformerL16(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            image_size=224,
            patch_size=16,
            num_layers=24,
            num_heads=16,
            hidden_dim=1024,
            mlp_dim=4096,
            **kwargs,
        )


class VisionTransformerL32(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            image_size=224,
            patch_size=32,
            num_layers=24,
            num_heads=16,
            hidden_dim=1024,
            mlp_dim=4096,
            **kwargs,
        )


class VisionTransformerH14(VisionTransformer):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            image_size=224,
            patch_size=14,
            num_layers=32,
            num_heads=16,
            hidden_dim=1280,
            mlp_dim=5120,
            **kwargs,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from pathlib import Path

import pandas as pd
import pytest

from renate.benchmark.experimentation import execute_experiment_job


@pytest.fixture
def experiment_job_kwargs():
    return {
        "backend": "local",
        "config_file": str(
            Path(__file__).parent.parent / "renate_config_files" / "config_scenario.py"
        ),
        "config_space": {"updater": "ER", "max_epochs": 5},
        "mode": "max",
        "metric": "val_accuracy",
        "num_updates": 2,
        "max_time": 15,
        "seed": 0,
    }


def test_execute_experiment_job(tmpdir, experiment_job_kwargs):
    """Only checking if things run, not testing anything besides that."""
    expected_columns = [
        "Task ID",
        "Average Accuracy",
        "Forgetting",
        "Forward Transfer",
        "Backward Transfer",
    ]
    expected_num_updates = experiment_job_kwargs["num_updates"]
    execute_experiment_job(experiment_outputs_url=tmpdir, **experiment_job_kwargs)
    results_df = pd.read_csv(str(Path(tmpdir) / "logs" / "metrics_summary.csv"))
    assert all(results_df.columns == expected_columns)
    for update_id in range(expected_num_updates):
        assert (Path(tmpdir) / f"update_{update_id}" / "learner.ckpt").is_file()
        assert (Path(tmpdir) / f"update_{update_id}" / "model.ckpt").is_file()
    assert (
        len(
            pd.read_csv(str(Path(tmpdir) / f"update_{expected_num_updates - 1}" / "hpo.csv"))[
                "update_id"
            ].unique()
        )
        == expected_num_updates
    )


@pytest.mark.parametrize(
    "update_dict,regex",
    [
        ({"backend": "UNKNOWN_BACKEND"}, r"Backend*"),
        ({"mode": "UNKNOWN_MODE"}, r"Mode*"),
        ({"num_updates": 3}, r"The dataset has*"),
    ],
    ids=["unknown backend", "unknown mode", "wrong number of updates"],
)
def test_execute_experiment_job_edge_cases(tmpdir, experiment_job_kwargs, update_dict, regex):
    """Check if basic input errors raise an exception."""
    experiment_job_kwargs.update(update_dict)
    with pytest.raises(AssertionError, match=regex):
        execute_experiment_job(experiment_outputs_url=tmpdir, **experiment_job_kwargs)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
import math
from typing import List, Optional

import torch
from torch.utils.data import Dataset, random_split

from renate import defaults
from renate.types import NestedTensors


def reinitialize_model_parameters(model: torch.nn.Module) -> None:
    """Reinitializes the parameters of a model.

    This relies on all submodules of `model` implementing a method `reset_parameters()`. This
    is implemented for core `torch.nn` layers, but this may not be the case for custom
    implementations of exotic layers. A warning is logged for modules that do not implement
    `reset_parameters()`.

    The actual logic of renitializing parameters depends on the type of layer. It may affect the
    module's buffers (non-trainable parameters, e.g., batch norm stats) as well.

    Args:
        model: The model to be re-initialized.
    """
    for module in model.modules():
        # Skip modules without any parameters of their own.
        if not list(module.parameters(recurse=False)) and not list(module.buffers(recurse=False)):
            continue
        try:
            module.reset_parameters()
        except AttributeError:
            logging.warning(f"Skipping module {module} while resetting parameters.")


def get_generator(seed: Optional[int] = None) -> torch.Generator:
    """Provides a torch.Generator for the given seed.

    torch.default_generator is returned if seed is None.
    """
    if seed is None:
        return torch.default_generator
    rng = torch.Generator(device="cpu")
    rng.manual_seed(seed)
    return rng


def randomly_split_data(
    dataset: Dataset, proportions: List[float], seed: int = defaults.SEED
) -> List[Dataset]:
    """Randomly splits a dataset into chunks."""
    rng = get_generator(seed)
    split_sizes = _proportions_into_sizes(proportions, len(dataset))
    return random_split(dataset, split_sizes, generator=rng)


def _proportions_into_sizes(proportions: List[float], size: int) -> List[int]:
    """A helper function to convert chunk proportions into sizes.

    In case of rounding doubts, any remaining samples are appended to the last split size.
    """
    assert math.isclose(sum(proportions), 1.0), sum(proportions)
    sizes = [round(proportion * size) for proportion in proportions[:-1]]
    sizes.append(size - sum(sizes))
    return sizes


def move_tensors_to_device(tensors: NestedTensors, device: torch.device) -> NestedTensors:
    """Moves a collection of tensors to `device`.

    The collection `tensors` can be a nested structure of tensors, tuples, lists, and dicts.
    """
    if isinstance(tensors, torch.Tensor):
        return tensors.to(device)
    elif isinstance(tensors, tuple):
        return tuple(move_tensors_to_device(t, device) for t in tensors)
    # We need to include lists here as well, since collate_fn sometimes turns tuples into lists.
    # See https://github.com/pytorch/pytorch/issues/48419.
    elif isinstance(tensors, list):
        return [move_tensors_to_device(t, device) for t in tensors]
    elif isinstance(tensors, dict):
        return {key: move_tensors_to_device(t, device) for key, t in tensors.items()}
    else:
        raise TypeError(
            "Expected `tensors` to be a nested structure of tensors, tuples, list and dict; "
            f"discovered {type(tensors)}."
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.defaults import TASK_ID
from renate.utils.optimizer import create_optimizer, create_scheduler


@pytest.mark.parametrize(
    "optimizer,kwargs",
    [
        ("SGD", {"lr": 0.01, "momentum": 0.0, "weight_decay": 0.5}),
        ("Adam", {"lr": 0.01, "weight_decay": 0.5}),
    ],
)
def test_make_optimizer_with_different_configurations(optimizer, kwargs):
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    params = model.get_params(task_id=TASK_ID)

    opt = create_optimizer(params, optimizer=optimizer, **kwargs)
    assert isinstance(opt, torch.optim.Optimizer)


def test_unknown_optimizer_raises_error():
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    params = model.get_params(task_id=TASK_ID)

    with pytest.raises(ValueError):
        create_optimizer(
            params, optimizer="UNKNOWN_OPTIMIZER", lr=0.01, momentum=0.0, weight_decay=0.5
        )


@pytest.mark.parametrize(
    "scheduler,kwargs",
    [
        ("ConstantLR", {}),
        ("ExponentialLR", {"gamma": 0.5}),
        ("StepLR", {"step_size": 10, "gamma": 0.5}),
    ],
)
def test_make_scheduler_with_different_configurations(scheduler, kwargs):
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    params = model.get_params(task_id=TASK_ID)
    opt = create_optimizer(params, optimizer="SGD", lr=0.01, momentum=0.0, weight_decay=0.5)

    sch = create_scheduler(opt, scheduler=scheduler, **kwargs)
    assert isinstance(sch, torch.optim.lr_scheduler._LRScheduler)


def test_unknown_scheduler_raises_error():
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    params = model.get_params(task_id=TASK_ID)
    opt = create_optimizer(params, optimizer="SGD", lr=0.01, momentum=0.0, weight_decay=0.5)

    with pytest.raises(ValueError):
        create_scheduler(opt, scheduler="UNKNOWN_SCHEDULER", gamma=0.5)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import List

import torch

import renate.defaults as defaults


def create_optimizer(
    params: List[torch.nn.Parameter],
    optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
    lr: float = defaults.LEARNING_RATE,
    momentum: float = defaults.MOMENTUM,
    weight_decay: float = defaults.WEIGHT_DECAY,
) -> torch.optim.Optimizer:
    """Creates optimizer used to train the model.

    Args:
        params: The list of parameters to be updated.
        optimizer: The name of the optimizer to be used. Currently 'Adam' and 'SGD' are supported.
        lr: Learning rate to be used.
        momentum: Value for the momentum hyperparameter (if relevant).
        weight_decay: Value for the weight_decay hyperparameter (if relevant).
    """

    if optimizer == "SGD":
        return torch.optim.SGD(params, lr, momentum, weight_decay)
    elif optimizer == "Adam":
        return torch.optim.Adam(params, lr, weight_decay=weight_decay)
    else:
        raise ValueError(f"Unknown optimizer: {optimizer}.")


def create_scheduler(
    optimizer: torch.optim.Optimizer,
    scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,
    step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
    gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
) -> torch.optim.lr_scheduler._LRScheduler:
    """Creates a learning rate scheduler used to train the model.

    Args:
        optimizer: The optimizer to be used.
        scheduler: The name of the scheduler to be used.
        step_size: Period of learning rate decay.
        gamma: Value for the gamma hyperparameter (if relevant).
    """

    if scheduler == "ConstantLR":
        return torch.optim.lr_scheduler.ConstantLR(optimizer, factor=1.0)
    elif scheduler == "ExponentialLR":
        return torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=gamma)
    elif scheduler == "StepLR":
        return torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma)
    else:
        raise ValueError(f"Unknown scheduler: {scheduler}.")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict, Optional

import torchmetrics

import renate.defaults as defaults


def create_metrics(
    task: defaults.SUPPORTED_TASKS_TYPE,
    additional_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
) -> torchmetrics.MetricCollection:
    """Creates task-specific metrics including all additional metrics.

    Args:
        task: Whether classification or regression, for now.
        additional_metrics: Dictionary of additionally metrics to be added to the returned
            `MetricCollection`.
    """
    if task == "classification":
        metric_collection = {
            "accuracy": torchmetrics.Accuracy(),
        }
    elif task == "regression":
        metric_collection = {"mean_squared_error": torchmetrics.MeanSquaredError()}
    else:
        raise NotImplementedError(f"Task {task} not implemented.")
    if additional_metrics:
        assert set(metric_collection).isdisjoint(set(additional_metrics)), (
            "Use a different name for your custom metrics. Following names are reserved for the "
            f"default metrics: {set(metric_collection)}."
        )
        metric_collection.update(additional_metrics)
    return torchmetrics.MetricCollection(metric_collection)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import copy

import pytest
import torch

import renate.defaults as defaults
from renate.updaters.experimental.joint import JointLearner


def get_model_and_dataset():
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    dataset = torch.utils.data.TensorDataset(
        torch.rand((100, 10)),
        torch.randint(10, (100,)),
    )
    return model, dataset


def test_joint_learner_memory_append():
    """This test checks that the memory buffer is updated correctly."""
    model, dataset = get_model_and_dataset()
    dataset_len = len(dataset)
    model_updater = pytest.helpers.get_simple_updater(
        model=model,
        learner_class=JointLearner,
        learner_kwargs={},
        max_epochs=1,
    )
    model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    assert len(model_updater._learner._memory_buffer) == dataset_len
    model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    assert len(model_updater._learner._memory_buffer) == 2 * dataset_len


def test_joint_learner_model_reset():
    """This test checks that the model is reinitialized correctly before an update."""
    model, dataset = get_model_and_dataset()
    model_updater = pytest.helpers.get_simple_updater(
        model=model,
        learner_class=JointLearner,
        learner_kwargs={"learning_rate": 0.0},
        max_epochs=1,
    )
    model = model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    model_copy = copy.deepcopy(model)
    model = model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    for (name, param), (name_copy, param_copy) in zip(
        model.named_parameters(), model_copy.named_parameters()
    ):
        assert name == name_copy
        assert not torch.allclose(param, param_copy)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch
import torchvision
from torch.utils.data import TensorDataset

from renate.utils import pytorch
from renate.utils.pytorch import randomly_split_data


@pytest.mark.parametrize("model", [torchvision.models.resnet18(pretrained=True)])
def test_reinitialize_model_parameters(model: torch.nn.Module):
    model.train()
    model(torch.randn(10, 3, 128, 128))  # Peform a forward pass, to change batchnorm buffers.
    params_before = [p.clone() for p in model.parameters()]
    buffers_before = [b.clone() for b in model.buffers()]
    pytorch.reinitialize_model_parameters(model)
    for p_before, p_after in zip(params_before, model.parameters()):
        assert not torch.allclose(p_before, p_after)
    for b_before, b_after in zip(buffers_before, model.buffers()):
        assert not torch.allclose(b_before, b_after)


@pytest.mark.parametrize(
    "proportions,expected_sizes",
    [
        [[0.5, 0.5], [50, 50]],
        [[0.3, 0.3, 0.4], [30, 30, 40]],
        [[0.333, 0.333, 0.334], [33, 33, 34]],
    ],
)
@pytest.mark.parametrize("splitter_func", [randomly_split_data])
def test_split_size(proportions, expected_sizes, splitter_func):
    dataset = TensorDataset(torch.randn(100, 1))
    splits = splitter_func(dataset, proportions, len(dataset))
    for i, split in enumerate(splits):
        assert len(split) == expected_sizes[i]


@pytest.mark.parametrize("splitter_func", [randomly_split_data])
def test_split_not_overlapping_splits(splitter_func):
    X = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    dataset = TensorDataset(X)
    splits = splitter_func(dataset, [0.5, 0.25, 0.25], len(dataset))
    for i in range(len(splits)):
        for j in range(len(splits)):
            if i == j:
                continue
            for k in range(len(splits[i])):
                assert splits[i][k] not in splits[j]


def test_random_splitting_sample_split_with_same_random_seed():
    X = torch.tensor([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    dataset = TensorDataset(X)

    d_1_split_1, d_1_split_2 = randomly_split_data(dataset, [0.5, 0.5], 42)
    d_2_split_1, d_2_split_2 = randomly_split_data(dataset, [0.5, 0.5], 42)

    for i in range(5):
        assert torch.equal(d_1_split_1[i][0], d_2_split_1[i][0])
        assert torch.equal(d_1_split_2[i][0], d_2_split_2[i][0])



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import importlib.util
import sys
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Union

import torchmetrics

from renate import defaults
from renate.benchmark.scenarios import Scenario
from renate.data.data_module import RenateDataModule
from renate.evaluation.evaluator import evaluate
from renate.models import RenateModule


def evaluate_and_record_results(
    results: Dict[str, List[List[float]]],
    model: RenateModule,
    data_module: Union[Scenario, RenateDataModule],
    transform: Optional[Callable] = None,
    target_transform: Optional[Callable] = None,
    task_id: str = defaults.TASK_ID,
    logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
    metric_postfix: str = "",
    batch_size: int = defaults.BATCH_SIZE,
    accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
    devices: Optional[int] = None,
) -> Dict[str, List[List[float]]]:
    """A helper function that performs the evaluation on test data and records quantitative metrics
    in a dictionary.

    Args:
        results: The results dictionary to which the results should be saved.
        model: A RenateModule to be evaluated.
        data_module: A Scenario or RenateDataModule from which the test data is queried.
        transform: The transformation applied for evaluation.
        target_transform: The target transformation applied for evaluation.
        task_id: The task ID for which the evaluation should be performed.
        logged_metrics: Metrics logged additional to the default ones.
        metric_postfix: The postfix for the metric names.
        batch_size: A batch size for the test loader.
        accelerator: Accelerator used by PyTorch Lightning to train the model.
        devices: Devices used by PyTorch Lightning to train the model. If the devices flag is not
            defined, it will assume devices to be "auto" and fetch the `auto_device_count` from the
            `accelerator`.
    """

    data_module.setup()

    update_results = evaluate(
        model=model,
        test_dataset=data_module.test_data(),
        task_id=task_id,
        batch_size=batch_size,
        transform=transform,
        target_transform=target_transform,
        logged_metrics=logged_metrics,
        accelerator=accelerator,
        devices=devices,
    )
    for key, value in update_results.items():
        if key not in results:
            results[key + metric_postfix] = []
        results[key + metric_postfix].append(value)
    return results


def get_model(config_module: ModuleType, **kwargs: Any) -> RenateModule:
    """Creates and returns a model instance."""
    return getattr(config_module, "model_fn")(**kwargs)


def get_data_module(config_module: ModuleType, **kwargs: Any) -> RenateDataModule:
    """Creates and returns a data module instance."""
    return getattr(config_module, "data_module_fn")(**kwargs)


def get_metrics(config_module: ModuleType) -> Dict[str, torchmetrics.Metric]:
    """Creates and returns a dictionary of metrics."""
    metrics_fn_name = "metrics_fn"
    if metrics_fn_name in vars(config_module):
        return getattr(config_module, metrics_fn_name)()


def get_and_prepare_data_module(config_module: ModuleType, **kwargs: Any) -> RenateDataModule:
    """Prepares data."""
    data_module = get_data_module(config_module, **kwargs)
    data_module.prepare_data()
    return data_module


def get_and_setup_data_module(
    config_module: ModuleType,
    prepare_data: bool,
    **kwargs: Any,
) -> RenateDataModule:
    """Creates data module and possibly calls the prepare_data function needed for setup"""
    data_module = get_data_module(config_module, **kwargs)
    if prepare_data:
        data_module.prepare_data()
    data_module.setup()
    return data_module


def import_module(module_name: str, location: str) -> ModuleType:
    """Imports Python module from file location."""
    spec = importlib.util.spec_from_file_location(name=module_name, location=location)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torchmetrics

from renate import defaults
from renate.evaluation.metrics.utils import create_metrics


@pytest.mark.parametrize(
    "task",
    defaults.SUPPORTED_TASKS_TYPE.__args__ + ("unsupported_task",),
    ids=defaults.SUPPORTED_TASKS_TYPE.__args__ + ("unsupported_task",),
)
def test_create_metrics_without_additional_metrics(task):
    """Tests all allowed cases and all edge cases when no additional metrics is provided.

    Edge cases: not supported task.
    """
    if task == "unsupported_task":
        with pytest.raises(NotImplementedError):
            create_metrics(task=task)
    else:
        metric_collection = create_metrics(task=task)
        assert metric_collection.prefix is None


@pytest.mark.parametrize(
    "additional_metrics, is_duplicate",
    [
        (
            {
                "auc": torchmetrics.AUC(),
                "calibration_error": torchmetrics.CalibrationError(),
            },
            False,
        ),
        (
            {
                "accuracy": torchmetrics.Accuracy(),
            },
            True,
        ),
    ],
    ids=["valid_metric_names", "duplicate_metric_names"],
)
def test_given_additional_metrics_create_metrics_adds_them_to_metric_collection(
    additional_metrics, is_duplicate
):
    """Passes if duplicate metric names raise an exception and additional metrics are added to the
    metric collection.

    Case 1: Valid case.
    Case 2: Metric name already used as part of the default metric collection.
    """
    if is_duplicate:
        with pytest.raises(AssertionError):
            create_metrics(task="classification", additional_metrics=additional_metrics)
    else:
        metric_collection = create_metrics(
            task="classification", additional_metrics=additional_metrics
        )
        assert metric_collection.prefix is None
        assert set(additional_metrics).issubset(set(metric_collection))



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import pytest
import torch

from renate.memory.storage import MemoryMappedTensorStorage


def nested_tensors_equal(t1, t2):
    if type(t1) is not type(t2):
        return False
    if isinstance(t1, torch.Tensor):
        return torch.equal(t1, t2)
    if isinstance(t1, tuple):
        if len(t1) != len(t2):
            return False
        return all(nested_tensors_equal(t1_, t2_) for t1_, t2_ in zip(t1, t2))
    if isinstance(t1, dict):
        if set(t1.keys()) != set(t2.keys()):
            return False
        return all(nested_tensors_equal(t1[key], t2[key]) for key in t1.keys())


@pytest.mark.parametrize("length", [0, 1, 10])
@pytest.mark.parametrize(
    "data_point",
    [
        torch.tensor(1),
        torch.ones(3),
        (torch.ones(3), torch.tensor(1)),
        ({"a": torch.ones(2, 3, 3), "b": torch.zeros(2)}, torch.tensor(4)),
        {"a": (torch.ones(2, 3, 3), torch.zeros(2)), "b": torch.tensor(2)},
    ],
)
def test_storage(tmpdir, length, data_point):
    """Tests the memory-mapped tensor storage for different nested tensor structures."""
    storage = MemoryMappedTensorStorage(tmpdir, data_point, length)
    for i in range(length):
        storage[i] = data_point

    del storage
    storage = MemoryMappedTensorStorage(tmpdir, data_point, length)

    for i in range(length):
        assert nested_tensors_equal(storage[i], data_point)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict, List


def average_accuracy(results: Dict[str, List[List[float]]], task_id: int) -> float:
    """Compute the average accuracy of a model.

    This measure is defined by:

    .. math::
        \\frac{1}{T}  sum_{i=1}^T  a_{T,i}

    where :math:`T` is the number of tasks, :math:`a_{T,i}` is the accuracy
    of the model on task :math:`i`, while having learned all tasks up to :math:`T`.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
    """
    return sum(results["accuracy"][task_id][: task_id + 1]) / (task_id + 1)


def forgetting(results: Dict[str, List[List[float]]], task_id: int) -> float:
    """Compute the forgetting measure of the model.

    This measure is defined by:

    .. math::
        \\frac{1}{T-1}  sum_{i=1}^{T-1}  f_{T,i}

    where :math:`f_{j,i}` is defined as:

    .. math::
        f_{j,i} = \\max_{k\\in\\{1, \\ldots, j=i\\}} a_{k,i} - a_{j,i}

    where :math:`T` is the final task index, :math:`a_{n,i}` is the test classification accuracy
    on task :math:`i` after sequentially learning the nth task and :math:`f_{j,i}` is a measure of
    forgetting on task :math:`i` after training up to task :math:`j`.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
    """
    if task_id == 0:
        return 0.0

    def f(results: Dict[str, List[List[float]]], j: int, i: int) -> float:
        """A Helper function to compute the: math:`f_{j,i}`."""
        accuracy_ji = results[j][i]
        max_accuracy_ki = max([results[k][i] for k in range(j)])
        return max_accuracy_ki - accuracy_ji

    sum_f = 0.0
    for i in range(task_id):
        sum_f += f(results["accuracy"], task_id, i)
    return sum_f / task_id


def backward_transfer(results: Dict[str, List[List[float]]], task_id: int) -> float:
    """Compute the backward transfer measure of the model.

    This measure is defined by:

    .. math::
        \\frac{1}{T-1}  sum_{i=1}^{T-1}  a_{T,i} - a_{i,i}

    where :math:`T` is the final task index, :math:`a_{n,i}` is the test classification accuracy
    on task :math:`i` after sequentially learning the nth task.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
    """
    if task_id == 0:
        return 0.0
    return (
        sum([results["accuracy"][task_id][i] - results["accuracy"][i][i] for i in range(task_id)])
        / task_id
    )


def forward_transfer(results: Dict[str, List[List[float]]], task_id: int) -> float:
    """Compute the forward transfer measure of the model.

    This measure is defined by:

    .. math::
        \\frac{1}{T-1}  sum_{i=2}^{T}  a_{i-1,i} - b_{i}

    where :math:`T` is the final task index, :math:`a_{n,i}` is the test classification accuracy
    on task :math:`i` after sequentially learning the nth task. :math:`b_{i}` is the accuracy for
    all :math:`T` tasks recorded at initialisation prior to observing any task.

    Args:
        results: The results dictionary holding all the results with respect to all recorded
            metrics.
        task_id: The task index.
    """
    if task_id == 0:
        return 0.0
    return sum(
        [
            results["accuracy"][i - 1][i] - results["accuracy_init"][0][i]
            for i in range(1, task_id + 1)
        ]
    ) / (task_id)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Type

import torch
import torchmetrics
from avalanche.training import Naive
from avalanche.training.plugins import LRSchedulerPlugin
from avalanche.training.supervised.icarl import _ICaRLPlugin
from avalanche.training.templates import BaseSGDTemplate
from syne_tune import Reporter
from torch.optim import Optimizer
from torch.utils.data import Dataset

from renate import defaults
from renate.models import RenateModule
from renate.updaters.avalanche.learner import (
    AvalancheEWCLearner,
    AvalancheICaRLLearner,
    AvalancheLwFLearner,
    AvalancheReplayLearner,
    ICaRL,
    plugin_by_class,
)
from renate.updaters.avalanche.plugins import (
    RenateCheckpointPlugin,
    RenateFileSystemCheckpointStorage,
)
from renate.updaters.learner import Learner
from renate.updaters.model_updater import SingleTrainingLoopUpdater
from renate.utils.avalanche import AvalancheBenchmarkWrapper, to_avalanche_dataset

logger = logging.getLogger(__name__)

metrics_mapper = {
    "train_loss": "Loss_Epoch/train_phase/train_stream/Task000",
    "train_accuracy": "Top1_Acc_Epoch/train_phase/train_stream/Task000",
    "val_loss": "Loss_Stream/eval_phase/test_stream/Task000",
    "val_accuracy": "Top1_Acc_Stream/eval_phase/test_stream/Task000",
}


class AvalancheModelUpdater(SingleTrainingLoopUpdater):
    _report = Reporter()

    def _load_learner(
        self,
        learner_class: Type[Learner],
        learner_kwargs: Dict[str, Any],
    ) -> BaseSGDTemplate:
        if self._early_stopping_enabled:  # TODO: support it
            raise Exception("Early stopping is not supported yet.")
        logger.warning(
            "Avalanche updaters currently support only accuracy and loss."
        )  # TODO: make use of passed metrics
        self._dummy_learner = learner_class(
            model=self._model,
            **learner_kwargs,
            logged_metrics=self._logged_metrics,
            **self._transforms_kwargs,
        )
        optimizer, scheduler = self._dummy_learner.configure_optimizers()
        optimizer, scheduler = optimizer[0], scheduler[0]
        lr_scheduler_plugin = LRSchedulerPlugin(scheduler=scheduler)
        plugins = [lr_scheduler_plugin]
        avalanche_learner = self._load_if_exists(self._input_state_folder)

        checkpoint_plugin = None
        if self._output_state_folder is not None:
            checkpoint_plugin = RenateCheckpointPlugin(
                RenateFileSystemCheckpointStorage(
                    directory=Path(self._output_state_folder),
                ),
            )
            plugins.append(checkpoint_plugin)

        if avalanche_learner is None:
            logger.warning("No updater state available. Updating from scratch.")
            return self._create_avalanche_learner(
                checkpoint_plugin=checkpoint_plugin,
                lr_scheduler_plugin=lr_scheduler_plugin,
                optimizer=optimizer,
            )

        self._dummy_learner.update_settings(
            avalanche_learner=avalanche_learner,
            plugins=plugins,
            optimizer=optimizer,
            max_epochs=self._max_epochs,
            device=self._get_device(),
            eval_every=1,
        )
        return avalanche_learner

    def _get_device(self) -> torch.device:
        """Returns the device according to the chosen accelerator."""
        if self._accelerator == "auto":
            return torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        elif self._accelerator == "gpu":
            return torch.device("cuda", self._devices or 0)
        elif self._accelerator == "tpu":
            raise Exception("Not supported accelerator `TPU`.")
        return torch.device("cpu")

    def _create_avalanche_learner(
        self,
        checkpoint_plugin: RenateCheckpointPlugin,
        lr_scheduler_plugin: LRSchedulerPlugin,
        optimizer: Optimizer,
    ) -> BaseSGDTemplate:
        """Returns an Avalanche learner based on the arguments passed to the ModelUpdater.

        Args:
            checkpoint_plugin: Plugin to checkpoint regularly.
            lr_scheduler_plugin: Plugin to adapt the learning rate.
            optimizer: PyTorch optimizer object used for training the Avalanche learner.
        """
        plugins = [lr_scheduler_plugin]
        if checkpoint_plugin is not None:
            plugins.append(checkpoint_plugin)
        avalanche_learner = self._dummy_learner.create_avalanche_learner(
            optimizer=optimizer,
            train_epochs=self._max_epochs,
            plugins=plugins,
            device=self._get_device(),
            eval_every=1,
        )
        avalanche_learner.is_logged_metric = (
            lambda metric_name: metric_name is None or metric_name in metrics_mapper.keys()
        )
        return avalanche_learner

    @staticmethod
    def _load_if_exists(input_state_folder: Optional[str]) -> Optional[Naive]:
        """Loads the Avalanche strategy if a state exists."""

        if input_state_folder is None:
            return None
        checkpoint_plugin = RenateCheckpointPlugin(
            RenateFileSystemCheckpointStorage(
                directory=Path(input_state_folder),
            ),
        )
        avalanche_learner, _ = checkpoint_plugin.load_checkpoint_if_exists()
        return avalanche_learner

    def update(
        self,
        train_dataset: Dataset,
        val_dataset: Optional[Dataset] = None,
        task_id: Optional[str] = None,
    ) -> RenateModule:
        val_dataset_exists = val_dataset is not None
        benchmark = self._load_benchmark_if_exists(train_dataset, val_dataset)
        train_exp = benchmark.train_stream[0]
        self._learner.train(train_exp, eval_streams=[benchmark.test_stream])
        results = self._learner.eval(benchmark.test_stream)
        if isinstance(self._learner, ICaRL):
            class_means = plugin_by_class(_ICaRLPlugin, self._learner.plugins).class_means
            self._model.class_means.data[:, :] = class_means
        if self._output_state_folder is not None:
            Path(self._output_state_folder).mkdir(
                exist_ok=True, parents=True
            )  # TODO: remove when checkpointing is active
            torch.save(self._model.state_dict(), defaults.model_file(self._output_state_folder))
            self._save_avalanche_state(benchmark, val_dataset_exists)
        self._report(
            **{
                metric_name: results[metric_internal_name]
                for metric_name, metric_internal_name in metrics_mapper.items()
                if val_dataset_exists or not metric_name.startswith("val")
            },
            step=self._max_epochs - 1,
            epoch=self._max_epochs,
        )
        return self._model

    def _load_benchmark_if_exists(
        self, train_dataset: Dataset, val_dataset: Optional[Dataset] = None
    ) -> AvalancheBenchmarkWrapper:
        train_dataset = to_avalanche_dataset(train_dataset)

        avalanche_state = None
        if self._input_state_folder is not None:
            avalanche_state_file = defaults.avalanche_state_file(self._input_state_folder)
            if Path(avalanche_state_file).exists():
                avalanche_state = torch.load(avalanche_state_file)
                if "val_memory_buffer" in avalanche_state:
                    self._dummy_learner._val_memory_buffer.load_state_dict(
                        avalanche_state["val_memory_buffer"]
                    )
                    self._dummy_learner.load(self._input_state_folder)
        if val_dataset is not None:
            self._dummy_learner._val_memory_buffer.update(val_dataset)
            val_memory_dataset = to_avalanche_dataset(self._dummy_learner._val_memory_buffer)
        else:
            val_memory_dataset = to_avalanche_dataset(train_dataset)

        benchmark = AvalancheBenchmarkWrapper(
            train_dataset=train_dataset,
            val_dataset=val_memory_dataset,
            train_transform=self._train_transform,
            train_target_transform=self._train_target_transform,
            test_transform=self._test_transform,
            test_target_transform=self._test_target_transform,
        )
        if avalanche_state is not None:
            benchmark.load_state_dict(avalanche_state)
        benchmark.update_benchmark_properties()
        return benchmark

    def _save_avalanche_state(self, benchmark: AvalancheBenchmarkWrapper, val_dataset_exists: bool):
        state = benchmark.state_dict()
        if val_dataset_exists:
            self._dummy_learner.save(self._output_state_folder)
            state["val_memory_buffer"] = self._dummy_learner._val_memory_buffer.state_dict()
        torch.save(state, defaults.avalanche_state_file(self._output_state_folder))


class ExperienceReplayAvalancheModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        memory_size: int,
        memory_batch_size: int = defaults.BATCH_SIZE,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "memory_batch_size": memory_batch_size,
            "optimizer": optimizer,
            "learning_rate": learning_rate,
            "learning_rate_scheduler": learning_rate_scheduler,
            "learning_rate_scheduler_gamma": learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": learning_rate_scheduler_step_size,
            "momentum": momentum,
            "weight_decay": weight_decay,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            learner_class=AvalancheReplayLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
        )


class ElasticWeightConsolidationModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        ewc_lambda: float = defaults.EWC_LAMBDA,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    ):
        learner_kwargs = {
            "optimizer": optimizer,
            "learning_rate": learning_rate,
            "learning_rate_scheduler": learning_rate_scheduler,
            "learning_rate_scheduler_gamma": learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": learning_rate_scheduler_step_size,
            "momentum": momentum,
            "weight_decay": weight_decay,
            "batch_size": batch_size,
            "ewc_lambda": ewc_lambda,
            "seed": seed,
        }
        super().__init__(
            model,
            learner_class=AvalancheEWCLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
        )


class LearningWithoutForgettingModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        alpha: float = defaults.LWF_ALPHA,
        temperature: float = defaults.LWF_TEMPERATURE,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    ):
        learner_kwargs = {
            "optimizer": optimizer,
            "learning_rate": learning_rate,
            "learning_rate_scheduler": learning_rate_scheduler,
            "learning_rate_scheduler_gamma": learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": learning_rate_scheduler_step_size,
            "momentum": momentum,
            "weight_decay": weight_decay,
            "batch_size": batch_size,
            "alpha": alpha,
            "temperature": temperature,
            "seed": seed,
        }
        super().__init__(
            model,
            learner_class=AvalancheLwFLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
        )


class ICaRLModelUpdater(AvalancheModelUpdater):
    def __init__(
        self,
        model: RenateModule,
        memory_size: int,
        memory_batch_size: int = defaults.BATCH_SIZE,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        buffer_transform: Optional[Callable] = None,
        buffer_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    ):
        learner_kwargs = {
            "memory_size": memory_size,
            "memory_batch_size": memory_batch_size,
            "optimizer": optimizer,
            "learning_rate": learning_rate,
            "learning_rate_scheduler": learning_rate_scheduler,
            "learning_rate_scheduler_gamma": learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": learning_rate_scheduler_step_size,
            "momentum": momentum,
            "weight_decay": weight_decay,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            learner_class=AvalancheICaRLLearner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            buffer_transform=buffer_transform,
            buffer_target_transform=buffer_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            accelerator=accelerator,
            devices=devices,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Callable, Dict, Optional

import torchmetrics
from pytorch_lightning.loggers.logger import Logger

from renate import defaults
from renate.models import RenateModule
from renate.updaters.learner import Learner
from renate.updaters.model_updater import SingleTrainingLoopUpdater


class FineTuningModelUpdater(SingleTrainingLoopUpdater):
    def __init__(
        self,
        model: RenateModule,
        optimizer: defaults.SUPPORTED_OPTIMIZERS_TYPE = defaults.OPTIMIZER,
        learning_rate: float = defaults.LEARNING_RATE,
        learning_rate_scheduler: defaults.SUPPORTED_LEARNING_RATE_SCHEDULERS_TYPE = defaults.LEARNING_RATE_SCHEDULER,  # noqa: E501
        learning_rate_scheduler_gamma: float = defaults.LEARNING_RATE_SCHEDULER_GAMMA,
        learning_rate_scheduler_step_size: int = defaults.LEARNING_RATE_SCHEDULER_STEP_SIZE,
        momentum: float = defaults.MOMENTUM,
        weight_decay: float = defaults.WEIGHT_DECAY,
        batch_size: int = defaults.BATCH_SIZE,
        input_state_folder: Optional[str] = None,
        output_state_folder: Optional[str] = None,
        max_epochs: int = defaults.MAX_EPOCHS,
        train_transform: Optional[Callable] = None,
        train_target_transform: Optional[Callable] = None,
        test_transform: Optional[Callable] = None,
        test_target_transform: Optional[Callable] = None,
        metric: Optional[str] = None,
        mode: defaults.SUPPORTED_TUNING_MODE_TYPE = "min",
        logged_metrics: Optional[Dict[str, torchmetrics.Metric]] = None,
        early_stopping_enabled: bool = False,
        logger: Logger = defaults.LOGGER(**defaults.LOGGER_KWARGS),
        accelerator: defaults.SUPPORTED_ACCELERATORS_TYPE = defaults.ACCELERATOR,
        devices: Optional[int] = None,
        seed: int = defaults.SEED,
        deterministic_trainer: bool = defaults.DETERMINISTIC_TRAINER,
    ):
        learner_kwargs = {
            "optimizer": optimizer,
            "learning_rate": learning_rate,
            "learning_rate_scheduler": learning_rate_scheduler,
            "learning_rate_scheduler_gamma": learning_rate_scheduler_gamma,
            "learning_rate_scheduler_step_size": learning_rate_scheduler_step_size,
            "momentum": momentum,
            "weight_decay": weight_decay,
            "batch_size": batch_size,
            "seed": seed,
        }
        super().__init__(
            model,
            learner_class=Learner,
            learner_kwargs=learner_kwargs,
            input_state_folder=input_state_folder,
            output_state_folder=output_state_folder,
            max_epochs=max_epochs,
            train_transform=train_transform,
            train_target_transform=train_target_transform,
            test_transform=test_transform,
            test_target_transform=test_target_transform,
            metric=metric,
            mode=mode,
            logged_metrics=logged_metrics,
            early_stopping_enabled=early_stopping_enabled,
            logger=logger,
            accelerator=accelerator,
            devices=devices,
            deterministic_trainer=deterministic_trainer,
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os

import pytest

from renate.utils.file import copy_to_uri, move_to_uri


def create_file(path, content):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w") as f:
        f.write(content)


@pytest.mark.parametrize(
    "file_content1_source_dir, file_content2_source_dir, file_content1_destination_dir,"
    "file_content2_destination_dir",
    [["test1_1", "test1_1", "test1_2", "test2_2"]],
)
def test_move_to_uri_locally(
    tmpdir,
    file_content1_source_dir,
    file_content2_source_dir,
    file_content1_destination_dir,
    file_content2_destination_dir,
):
    """Test for moving files from a local directory to another local directory.

    The files should be moved from the source directory and the destination directory should be
    created if it does not exist.

    If there are files with the same name in the destination directory as in the source directory
    they should be overwritten.
    """

    create_file(os.path.join(tmpdir, "source_dir", "file1.txt"), file_content1_source_dir)
    create_file(os.path.join(tmpdir, "source_dir", "file2.txt"), file_content2_source_dir)

    create_file(os.path.join(tmpdir, "destination_dir", "file1.txt"), file_content1_destination_dir)
    create_file(os.path.join(tmpdir, "destination_dir", "file3.txt"), file_content2_destination_dir)

    move_to_uri(os.path.join(tmpdir, "source_dir"), os.path.join(tmpdir, "destination_dir"))

    with open(os.path.join(tmpdir, "destination_dir", "file1.txt"), "r") as f:
        assert f.read() == file_content1_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file2.txt"), "r") as f:
        assert f.read() == file_content2_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file3.txt"), "r") as f:
        assert f.read() == file_content2_destination_dir

    assert not os.path.exists(os.path.join(tmpdir, "source_dir", "file1.txt"))
    assert not os.path.exists(os.path.join(tmpdir, "source_dir", "file2.txt"))


@pytest.mark.parametrize(
    "file_content1_source_dir, file_content2_source_dir, file_content1_destination_dir, "
    "file_content2_destination_dir",
    [["test1_1", "test1_1", "test1_2", "test2_2"]],
)
def test_copy_to_uri_locally(
    tmpdir,
    file_content1_source_dir,
    file_content2_source_dir,
    file_content1_destination_dir,
    file_content2_destination_dir,
):
    """Test for copying files from a local directory to another local directory.

    The files should be copied from the source directory and the destination directory should be
    created if it does not exist.

    If there are files with the same name in the destination directory as in the source directory
    they should be overwritten.

    The source directory should not be changed.
    """

    create_file(os.path.join(tmpdir, "source_dir", "file1.txt"), file_content1_source_dir)
    create_file(os.path.join(tmpdir, "source_dir", "file2.txt"), file_content2_source_dir)

    create_file(os.path.join(tmpdir, "destination_dir", "file1.txt"), file_content1_destination_dir)
    create_file(os.path.join(tmpdir, "destination_dir", "file3.txt"), file_content2_destination_dir)

    copy_to_uri(os.path.join(tmpdir, "source_dir"), os.path.join(tmpdir, "destination_dir"))

    with open(os.path.join(tmpdir, "destination_dir", "file1.txt"), "r") as f:
        assert f.read() == file_content1_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file2.txt"), "r") as f:
        assert f.read() == file_content2_source_dir
    with open(os.path.join(tmpdir, "destination_dir", "file3.txt"), "r") as f:
        assert f.read() == file_content2_destination_dir

    with open(os.path.join(tmpdir, "source_dir", "file1.txt"), "r") as f:
        assert f.read() == file_content1_source_dir
    with open(os.path.join(tmpdir, "source_dir", "file2.txt"), "r") as f:
        assert f.read() == file_content2_source_dir



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os

import pytest
import torch

from renate import defaults
from renate.updaters.experimental.er import (
    CLSExperienceReplayLearner,
    DarkExperienceReplayLearner,
    ExperienceReplayLearner,
    PooledOutputDistillationExperienceReplayLearner,
    SuperExperienceReplayLearner,
)


def get_model_and_dataset():
    model = pytest.helpers.get_renate_module_mlp(
        num_outputs=10, num_inputs=10, hidden_size=32, num_hidden_layers=3
    )
    dataset = torch.utils.data.TensorDataset(
        torch.rand((100, 10)),
        torch.randint(10, (100,)),
    )
    return model, dataset


@pytest.mark.parametrize(
    "batch_size,memory_size,memory_batch_size",
    [[10, 10, 10], [20, 10, 10], [10, 100, 10], [10, 30, 1], [100, 10, 3]],
)
def test_er_overall_memory_size_after_update(batch_size, memory_size, memory_batch_size):
    model, dataset = get_model_and_dataset()
    learner_kwargs = {
        "memory_size": memory_size,
        "memory_batch_size": memory_batch_size,
        "batch_size": batch_size,
    }
    model_updater = pytest.helpers.get_simple_updater(
        model=model,
        learner_class=ExperienceReplayLearner,
        learner_kwargs=learner_kwargs,
        max_epochs=1,
    )
    model_updater.update(train_dataset=dataset, task_id=defaults.TASK_ID)
    memory, _ = next(iter(model_updater._learner._memory_loader))
    x, y = memory
    assert x.shape[0] == memory_batch_size and y.shape[0] == memory_batch_size
    assert len(model_updater._learner._memory_buffer) == memory_size


def test_er_validation_buffer(tmpdir):
    model, dataset_train, _ = pytest.helpers.get_renate_module_mlp_and_data(
        num_inputs=10,
        num_outputs=10,
        hidden_size=32,
        num_hidden_layers=3,
        train_num_samples=10,
        test_num_samples=5,
    )
    datasets_val = []
    state_folder = None
    next_state_folder = defaults.input_state_folder(tmpdir)
    for i in range(3):
        dataset_val = torch.utils.data.TensorDataset(
            torch.rand((100, 10)),
            torch.randint(10, (100,)),
        )
        model_updater = pytest.helpers.get_simple_updater(
            model, input_state_folder=state_folder, output_state_folder=next_state_folder
        )
        model_updater.update(
            train_dataset=dataset_train, val_dataset=dataset_val, task_id=defaults.TASK_ID
        )
        datasets_val.append(dataset_val)
        state_folder = next_state_folder

    model_updater = pytest.helpers.get_simple_updater(
        model, input_state_folder=state_folder, output_state_folder=next_state_folder
    )
    for i in range(3):
        for j in range(100):
            assert torch.allclose(
                datasets_val[i][j][0],
                model_updater._learner._val_memory_buffer[i * 100 + j][0][0],
                rtol=1e-3,
            )


@pytest.mark.parametrize(
    "cls,kwargs",
    [
        [ExperienceReplayLearner, {"alpha": 0.2, "memory_size": 10, "memory_batch_size": 10}],
        [DarkExperienceReplayLearner, {"alpha": 0.1, "beta": 0.3, "memory_size": 42}],
        [
            CLSExperienceReplayLearner,
            {
                "alpha": 0.3,
                "beta": 0.1,
                "stable_model_update_weight": 0.3,
                "plastic_model_update_weight": 0.3,
                "stable_model_update_probability": 0.3,
                "plastic_model_update_probability": 0.5,
                "memory_size": 42,
            },
        ],
        [
            PooledOutputDistillationExperienceReplayLearner,
            {"alpha": 0.3, "distillation_type": "pixel", "normalize": False, "memory_size": 42},
        ],
        [
            SuperExperienceReplayLearner,
            {
                "der_alpha": 0.2,
                "der_beta": 0.3,
                "sp_shrink_factor": 0.1,
                "sp_sigma": 0.3,
                "cls_alpha": 0.3,
                "cls_stable_model_update_weight": 0.4,
                "cls_plastic_model_update_weight": 0.4,
                "cls_stable_model_update_probability": 0.3,
                "cls_plastic_model_update_probability": 0.5,
                "pod_alpha": 0.1,
                "pod_distillation_type": "pixel",
                "pod_normalize": False,
                "memory_size": 42,
            },
        ],
    ],
)
def test_er_components_save_and_load(tmpdir, cls, kwargs):
    """This test saves the learner state and reloads it and verifies that the hyperparameters
    for the components were correctly set, saved and loaded."""
    model = pytest.helpers.get_renate_module_mlp(
        num_inputs=10, num_outputs=10, hidden_size=32, num_hidden_layers=3
    )
    learner = cls(model=model, **kwargs)
    torch.save(learner.state_dict(), os.path.join(tmpdir, "learner.pt"))
    learner = cls.__new__(cls)
    learner.load_state_dict(model, torch.load(os.path.join(tmpdir, "learner.pt")))
    if isinstance(learner, ExperienceReplayLearner) and not isinstance(
        learner, DarkExperienceReplayLearner
    ):
        assert learner._components["memory_loss"]._weight == kwargs["alpha"]
    if isinstance(learner, DarkExperienceReplayLearner):
        assert learner._components["mse_loss"]._weight == kwargs["alpha"]
        assert learner._components["memory_loss"]._weight == kwargs["beta"]
    elif isinstance(learner, PooledOutputDistillationExperienceReplayLearner):
        assert learner._components["pod_loss"]._weight == kwargs["alpha"]
        assert learner._components["pod_loss"]._distillation_type == kwargs["distillation_type"]
        assert learner._components["pod_loss"]._normalize == kwargs["normalize"]
    elif isinstance(learner, CLSExperienceReplayLearner):
        assert learner._components["memory_loss"]._weight == kwargs["alpha"]
        assert learner._components["cls_loss"]._weight == kwargs["beta"]
        assert (
            learner._components["cls_loss"]._stable_model_update_weight
            == kwargs["stable_model_update_weight"]
        )
        assert (
            learner._components["cls_loss"]._plastic_model_update_weight
            == kwargs["plastic_model_update_weight"]
        )
        assert (
            learner._components["cls_loss"]._stable_model_update_probability
            == kwargs["stable_model_update_probability"]
        )
        assert (
            learner._components["cls_loss"]._plastic_model_update_probability
            == kwargs["plastic_model_update_probability"]
        )
    elif isinstance(learner, SuperExperienceReplayLearner):
        assert learner._components["mse_loss"]._weight == kwargs["der_alpha"]
        assert learner._components["memory_loss"]._weight == kwargs["der_beta"]
        assert learner._components["shrink_perturb"]._shrink_factor == kwargs["sp_shrink_factor"]
        assert learner._components["shrink_perturb"]._sigma == kwargs["sp_sigma"]
        assert learner._components["cls_loss"]._weight == kwargs["cls_alpha"]
        assert (
            learner._components["cls_loss"]._stable_model_update_weight
            == kwargs["cls_stable_model_update_weight"]
        )
        assert (
            learner._components["cls_loss"]._plastic_model_update_weight
            == kwargs["cls_plastic_model_update_weight"]
        )
        assert (
            learner._components["cls_loss"]._stable_model_update_probability
            == kwargs["cls_stable_model_update_probability"]
        )
        assert (
            learner._components["cls_loss"]._plastic_model_update_probability
            == kwargs["cls_plastic_model_update_probability"]
        )
        assert learner._components["pod_loss"]._weight == kwargs["pod_alpha"]
        assert learner._components["pod_loss"]._distillation_type == kwargs["pod_distillation_type"]
        assert learner._components["pod_loss"]._normalize == kwargs["pod_normalize"]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Callable, Optional, Tuple
from typing import List

import torch
from PIL import Image
from torch import Tensor
from torch.utils.data import Dataset

from renate.types import NestedTensors


class ImageDataset(Dataset):
    """Dataset class for image datasets where the images are loaded as raw images.

    Args:
        data: List of data paths to the images.
        labels: Labels of images.
        transform: Transformation or augmentation to perform on the sample.
        target_transform: Transformation or augmentation to perform on the target.
    """

    def __init__(
        self,
        data: List[str],
        labels: List[int],
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        assert len(data) == len(labels)
        self.data = data
        self.labels = labels
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx) -> Tuple[Tensor, Tensor]:
        with open(self.data[idx], "rb") as f:
            image = Image.open(f)
            image.load()

        # In case of black and white images, we should convert them to rgb
        if len(image.size) <= 2:
            image = image.convert("RGB")

        target = torch.tensor(self.labels[idx], dtype=torch.long)

        if self.transform:
            image = self.transform(image)

        if self.target_transform:
            target = self.target_transform(target)
        return image, target.long()


class NestedTensorDataset(Dataset):
    """A dataset of nested tensors.

    Args:
        nested_tensors: A nested tuple/dict structure of tensors. Tensors need to be of equal size
            along the batch dimension.
    """

    def __init__(self, nested_tensors: NestedTensors):
        self._nested_tensors = nested_tensors
        self._length = self._get_len(nested_tensors)

    def _get_len(self, nested_tensors: NestedTensors, expected_length: Optional[int] = None) -> int:
        """Extracts length of the dataset and checks for consistent length."""
        if isinstance(nested_tensors, torch.Tensor):
            length = nested_tensors.size(0)
            assert length == expected_length or expected_length is None
            return length
        elif isinstance(nested_tensors, tuple):
            for t in nested_tensors:
                expected_length = self._get_len(t, expected_length)
            return expected_length
        elif isinstance(nested_tensors, dict):
            for t in nested_tensors.values():
                expected_length = self._get_len(t, expected_length)
            return expected_length
        else:
            raise TypeError(f"Expected nested dict/tuple of tensors, found {type(nested_tensors)}.")

    def __len__(self):
        """Returns the number of data points in the dataset."""
        return self._length

    @staticmethod
    def _get(storage: NestedTensors, idx: int) -> NestedTensors:
        if isinstance(storage, torch.Tensor):
            return storage[idx]
        elif isinstance(storage, tuple):
            return tuple(NestedTensorDataset._get(t, idx) for t in storage)
        elif isinstance(storage, dict):
            return {key: NestedTensorDataset._get(t, idx) for key, t in storage.items()}
        else:
            raise TypeError(f"Expected nested tuple/dict of tensors, found {type(storage)}.")

    def __getitem__(self, idx: int) -> NestedTensors:
        """Read the item stored at index `idx`."""
        return self._get(self._nested_tensors, idx)


class _TransformedDataset(Dataset):
    """A dataset wrapper that applies transformations.

    This wrapper assumes that the the passed `dataset` returns a pair `(x, y)`.

    Args:
        dataset: The dataset to wrap.
        transform: Transformation to perform on the covariates `x`.
        target_transform: Transformation to perform on the target `y`.
    """

    def __init__(
        self,
        dataset: Dataset,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
    ) -> None:
        super(Dataset, self).__init__()
        self._dataset = dataset
        self._transform = transform
        self._target_transform = target_transform

    def __getitem__(self, idx: int) -> Tuple[Tensor, Tensor]:
        """Returns the transformed data point."""
        data, target = self._dataset[idx]
        if self._transform is not None:
            data = self._transform(data)
        if self._target_transform is not None:
            target = self._target_transform(target)
        return data, target

    def __len__(self) -> int:
        """Returns the number of data points in the dataset."""
        return len(self._dataset)


class _EnumeratedDataset(Dataset):
    """A dataset wrapper that enumerates data points.

    Args:
        dataset: The dataset to wrap.
    """

    def __init__(self, dataset: Dataset) -> None:
        super().__init__()
        self._dataset = dataset

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, Any]:
        """Returns index and data point."""
        return torch.tensor(idx, dtype=torch.long), self._dataset[idx]

    def __len__(self) -> int:
        """Returns the number of data points in the dataset."""
        return len(self._dataset)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import Dict, Optional

import torch
import torch.nn as nn
from torch import Tensor
from transformers import AutoModelForSequenceClassification

from renate.models import RenateModule


class HuggingFaceSequenceClassificationTransformer(RenateModule):
    """RenateModule which wraps around Hugging Face transformers.

    Args:
        pretrained_model_name: Hugging Face model id.
        num_outputs: Number of outputs.
        loss_fn: The loss function to be optimized during the training.
    """

    def __init__(
        self,
        pretrained_model_name: str,
        num_outputs: int,
        loss_fn: nn.Module = nn.CrossEntropyLoss(),
    ) -> None:
        super().__init__(
            constructor_arguments={
                "pretrained_model_name": pretrained_model_name,
                "num_outputs": num_outputs,
            },
            loss_fn=loss_fn,
        )
        self._model = AutoModelForSequenceClassification.from_pretrained(
            pretrained_model_name, num_labels=num_outputs, return_dict=False
        )

    def forward(self, x: Dict[str, Tensor], task_id: Optional[str] = None) -> torch.Tensor:
        return self._model(**x)[0]

    def _add_task_params(self, task_id: str) -> None:
        assert not len(self._tasks_params_ids), "Transformer does not work for multiple tasks."



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os
from typing import Any, Dict, Tuple, Type

import pytest
import torch
from torchvision.transforms import ToTensor

from conftest import (
    LEARNERS,
    LEARNER_HYPERPARAMETER_UPDATES,
    LEARNER_KWARGS,
    check_learner_transforms,
)
from renate.models import RenateModule
from renate.updaters.learner import Learner, ReplayLearner


def get_model_and_learner_and_learner_kwargs(
    learner_class: Type[Learner],
) -> Tuple[RenateModule, Learner, Dict[str, Any]]:
    learner_kwargs = LEARNER_KWARGS[learner_class]
    model = pytest.helpers.get_renate_module_mlp(
        num_inputs=1, num_outputs=1, hidden_size=1, num_hidden_layers=1
    )
    learner = learner_class(model=model, **learner_kwargs)
    return model, learner, learner_kwargs


def check_learner_variables(learner: Learner, expected_variable_values: Dict[str, Any]):
    for attribute_name, attribute_value in expected_variable_values.items():
        if attribute_name == "memory_size":
            continue
        assert getattr(learner, f"_{attribute_name}") == attribute_value


@pytest.mark.parametrize("learner_class", LEARNERS)
def test_save_and_load_learner(tmpdir, learner_class):
    model, learner, learner_kwargs = get_model_and_learner_and_learner_kwargs(learner_class)
    filename = os.path.join(tmpdir, "learner.pkl")
    torch.save(learner.state_dict(), filename)
    learner = learner_class.__new__(learner_class)
    learner.load_state_dict(model, torch.load(filename))
    check_learner_variables(learner, learner_kwargs)


@pytest.mark.parametrize("learner_class", LEARNERS)
def test_update_hyperparameters(learner_class):
    model, learner, learner_kwargs = get_model_and_learner_and_learner_kwargs(learner_class)
    check_learner_variables(learner, learner_kwargs)
    learner.update_hyperparameters({})
    check_learner_variables(learner, learner_kwargs)
    learner.update_hyperparameters(LEARNER_HYPERPARAMETER_UPDATES[learner_class])
    learner_kwargs = dict(learner_kwargs)
    learner_kwargs.update(LEARNER_HYPERPARAMETER_UPDATES[learner_class])
    check_learner_variables(learner, learner_kwargs)


@pytest.mark.parametrize("learner_class", LEARNERS)
def test_set_transforms(learner_class):
    """Tests if set_transforms function correctly sets transforms in Learner and MemoryBuffer."""
    model, learner, learner_kwargs = get_model_and_learner_and_learner_kwargs(learner_class)
    check_learner_transforms(learner, {})
    transforms_kwargs = {
        "train_transform": ToTensor(),
        "train_target_transform": ToTensor(),
        "test_transform": ToTensor(),
        "test_target_transform": ToTensor(),
    }
    if issubclass(learner_class, ReplayLearner):
        transforms_kwargs["buffer_transform"] = ToTensor()
        transforms_kwargs["buffer_target_transform"] = ToTensor()
    learner.set_transforms(**transforms_kwargs)
    check_learner_transforms(learner, transforms_kwargs)
